import pycomm3

from ..types import *
from . import helper
from . import paths
from . import registry

# Known RemoteHelper file version numbers, used to help check that device is a valid terminal.
HELPER_VERSIONS = {
    '5.10.00',
    '7.00.00',
    '8.00.00',
    '8.10.00',
    '8.20.00',
    '9.00.00',
    '10.00.00',
    '11.00.00',
    '12.00.00',
    '12.00.00.414.73',
    '13.00.00'
}

# Known terminal MEVersion numbers, used to help check that device is a valid terminal.
ME_VERSIONS = {
    '5.10.16.09',
    '7.00.20.13',
    '7.00.55.13',
    '8.00.67.12',
    '8.10.42.13',
    '8.20.30.10',
    '9.00.17.241',
    '9.00.00.241',
    '10.00.09.290',
    '11.00.00.230',
    '11.00.25.230',
    '12.00.00.414',
    '12.00.78.414',
    '13.00.11.413'
}

# Known terminal product codes, used to help check that device is a valid terminal.
PRODUCT_CODES = {
    '17',   #PanelView Plus
    '47',   #PanelView Plus 6 1000
    '48',   #PanelView Plus 6
    '51',   #PanelView Plus 6
    '94',   #PanelView Plus 7 700 Perf
    '98',   #PanelView Plus 7 1000 Perf
    '102',  #PanelView Plus 7
    '187',  #PanelView Plus 7 1000 Standard
    '189'   #PanelView Plus 7 1200 Standard
}

# Known product types, used to help check that device is a valid terminal.
PRODUCT_TYPES = {
    '24'
}

def get_terminal_info(cip: pycomm3.CIPDriver) -> MEDeviceInfo:
    me_version = registry.get_me_version(cip)
    major_rev = int(me_version.split(".")[0])

    if major_rev <= 5:
        paths.helper_path = '\\Storage Card\\Rockwell Software\\RSViewME'
        paths.helper_file_path = paths.helper_path + '\\' + paths.HELPER_FILE_NAME
        paths.storage_path = '\\Storage Card'
        paths.upload_list_path = paths.storage_path + '\\' + paths.UPLOAD_LIST_PATH

    return MEDeviceInfo(cip._cip_path, 
                        helper.get_helper_version(cip),
                        me_version,
                        registry.get_product_code(cip),
                        registry.get_product_type(cip),
                        [],
                        [])

def is_terminal_valid(device: MEDeviceInfo) -> bool:
    if device.helper_version not in HELPER_VERSIONS: return False
    if device.me_version not in ME_VERSIONS: return False
    if device.product_code not in PRODUCT_CODES: return False
    if device.product_type not in PRODUCT_TYPES: return False
    return True

def is_download_valid(cip: pycomm3.CIPDriver, device: MEDeviceInfo, file: MEFile) -> bool:
    # Check that file is correct extension
    if (file.get_ext() != '.mer'):
        device.log.append(f'File {file.name} is not a *.mer file')
        return False

    # Check that storage folder exists
    resp_storage_exists = helper.get_folder_exists(cip)
    if not(resp_storage_exists):
        device.log.append(f'Storage folder does not exist on terminal')
        return False

    # Check free space
    resp_free_space = helper.get_free_space(cip)
    if (resp_free_space > file.get_size()):
        device.log.append(f'File {file.name} requires {file.get_size()} byes.  Free space on terminal {resp_free_space} bytes.')
    else:
        device.log.append(f'File {file.name} requires {file.get_size()} bytes.  Free space on terminal {resp_free_space} bytes is insufficient.')
        return False

    # Check if file name already exists
    resp_file_exists = helper.get_file_exists(cip, file)
    file.overwrite_required = False
    if (resp_file_exists and file.overwrite_requested):
        device.log.append(f'File {file.name} already exists on terminal, and overwrite was requested.  Setting overwrite to required.')
        file.overwrite_required = True
    if (resp_file_exists and not(file.overwrite_requested)):
        device.log.append(f'File {file.name} already exists on terminal, and overwrite was NOT requested.  Use kwarg overwrite_requested=True to overwrite existing.')
        return False
    if not(resp_file_exists):
        device.log.append(f'File {file.name} does not exist on terminal.  Setting overwrite to not required.')

    # Check space consumed by file if it exists
    if resp_file_exists:
        resp_file_size = helper.get_file_size(cip, file)
        device.log.append(f'File {file.name} on terminal is {resp_file_size} bytes.')

    return True


import pycomm3

from ..types import *
from . import files
from . import helper
from . import paths
from . import registry

def create_log(cip: pycomm3.CIPDriver, device: MEDeviceInfo):
    device.log.append(f'Terminal storage exists: {helper.get_folder_exists(cip)}.')
    device.log.append(f'Terminal has {helper.get_free_space(cip)} free bytes')
    device.log.append(f'Terminal has files: {upload_mer_list(cip, device)}')
    device.log.append(f'Terminal startup file: {registry.get_startup_mer(cip)}.')

def download_mer_file(cip: pycomm3.CIPDriver, device: MEDeviceInfo, file: MEFile, run_at_startup: bool, replace_comms: bool, delete_logs: bool) -> bool:
    # Create runtime folder
    #
    # TODO: Can we check if this already exists and skip?
    if not(helper.create_runtime_directory(cip, file)): raise Exception('Failed to create runtime path on terminal.')

    # Get attributes
    #
    # Still no clue on what these are, or when/how they would change.
    # If they aren't changed by creating paths, could be moved ahead
    # to is_download_valid().
    if not(files.is_get_unk_valid(cip)): raise Exception('Invalid response from an unknown attribute.  Check packets.')

    # Create a file exchange on the terminal
    file_instance = files.create_exchange_download(cip, file, paths.storage_path + '\\Rockwell Software\\RSViewME\\Runtime')
    device.log.append(f'Create file exchange {file_instance} for download.')

    # Set attributes
    #
    # Still no clue what this is.  Might be setting file up for write?
    if not(files.is_set_unk_valid(cip)): raise Exception('Invalid response from an unknown attribute.  Check packets.')

    # Transfer *.MER chunk by chunk
    files.download(cip, file_instance, file.path)

    # Mark file exchange as completed on the terminal
    files.end_write(cip, file_instance)
    device.log.append(f'Downloaded {file.path} to {file.name} using file exchange {file_instance}.')

    # Delete file exchange on the terminal
    files.delete_exchange(cip, file_instance)
    device.log.append(f'Deleted file exchange {file_instance}.')

    # Set *.MER to run at startup and then reboot
    if run_at_startup:
        helper.set_startup_mer(cip, file, replace_comms, delete_logs)
        device.log.append(f'Setting file to run at startup.')
        reboot(cip)
        device.log.append(f'Rebooting terminal.')

    return True

def upload_mer_file(cip: pycomm3.CIPDriver, device: MEDeviceInfo, file: MEFile, rem_file: MEFile) -> bool:
    # Verify file exists on terminal
    if not(helper.get_file_exists(cip, rem_file)): raise Exception(f'File {rem_file.name} does not exist on terminal.')

    # Create file exchange
    file_instance = files.create_exchange_upload(cip, paths.storage_path + f'\\Rockwell Software\\RSViewME\\Runtime\\{rem_file.name}')
    device.log.append(f'Create file exchange {file_instance} for upload.')

    # Transfer *.MER chunk by chunk
    files.upload_mer(cip, file_instance, file)
    device.log.append(f'Uploaded {rem_file.name} to {file.path} using file exchange {file_instance}.')

    # Delete file exchange on the terminal
    files.delete_exchange(cip, file_instance)
    device.log.append(f'Deleted file exchange {file_instance}.')

    return True

def upload_mer_list(cip: pycomm3.CIPDriver, device: MEDeviceInfo):
    # Create *.MER list
    helper.create_mer_list(cip)

    # Create file exchange on the terminal
    file_instance = files.create_exchange_upload(cip, paths.upload_list_path)
    device.log.append(f'Create file exchange {file_instance} for upload.')

    # Transfer *.MER list chunk by chunk
    file_list = files.upload_mer_list(cip, file_instance)
    device.log.append(f'Uploaded *.MER list using file exchange {file_instance}.')
    device.files = file_list

    # Delete file exchange on the terminal
    files.delete_exchange(cip, file_instance)
    device.log.append(f'Deleted file exchange {file_instance}.')

    # Delete *.MER list on the terminal
    helper.delete_file_mer_list(cip)
    device.log.append(f'Delete *.MER list on terminal.')

    return file_list

def reboot(cip: pycomm3.CIPDriver):
    cip = pycomm3.CIPDriver(cip._cip_path)
    cip._cfg['socket_timeout'] = 0.25
    cip.open()
    helper.reboot(cip)
    cip.close()


HELPER_FILE_NAME = 'RemoteHelper.DLL'
UPLOAD_LIST_PATH = 'Rockwell Software\\RSViewME\\Runtime\\Results.txt'

helper_path = '\\Windows'
helper_file_path = helper_path + '\\' + HELPER_FILE_NAME

storage_path = '\\Application Data'
upload_list_path = storage_path + '\\' + UPLOAD_LIST_PATH


from enum import Enum
import pycomm3
import struct
from warnings import warn

from .. import messages
from ..types import *
from .paths import *

# Known static value for successful creation of a folder.
# Further investigation needed.
CREATE_DIR_SUCCESS = 183 # 'b\xb7'

# Known functions available from RemoteHelper.
class HelperFunctions(Enum):
    REBOOT = 'BootTerminal'
    # ex: 'BootTerminal',''
    #       Returns a comms error since upon successful execution, terminal is reset.

    CREATE_DIRECTORY = 'CreateRemDirectory'
    # ex: 'CreateRemDirectory','{Path}'
    #       Returns a static value if successful, further investigation needed.

    CREATE_SHORTCUT = 'CreateRemShortcut'

    CREATE_ME_SHORTCUT = 'CreateRemMEStartupShortcut'
    # ex: 'CreateRemMEStartupShortcut','{Folder Path}:{Filename.MER}: /r /delay'
    #       Can also specify /o to replace comms and /d to delete logs at startup
    #       Returns 1 if shortcut created successfully, 0 otherwise.

    DELETE_DIRECTORY = 'DeleteRemDirectory'

    DELETE_FILE = 'DeleteRemFile'

    CREATE_FILE_LIST = 'FileBrowse'
    # ex: 'FileBrowse','{Search Path}\\*.{Search Extension}::{Results File Path}
    #       Returns 1 if results file generated successfully, 0 otherwise.

    GET_FILE_EXISTS = 'FileExists'
    # ex: 'FileExists','{File Path}'
    #       Returns 1 if file exists, 0 otherwise.

    GET_FILE_SIZE = 'FileSize'
    # ex: 'FileSize','{File Path}'
    #       Returns the number of bytes.

    CREATE_FOLDER_LIST = 'FolderBrowse'

    GET_FREE_SPACE = 'FreeSpace'
    # ex: 'FreeSpace','{Path}'
    #       Returns the number of bytes.
    GET_FILE_VERSION = 'GetFileVersion'
    GET_VERSION = 'GetVersion'
    # ex: 'GetVersion','{File Path}'
    #       Returns the a version string.

    START_PROCESS = 'InvokeEXE'
    GET_EXE_RUNNING = 'IsExeRunning'
    GET_FOLDER_EXISTS = 'StorageExists'
    # ex: 'StorageExists'.'{Path}'
    #       Returns 1 if folder exists, 0 otherwise.
    STOP_PROCESS = 'TerminateEXE'

def run_function(cip: pycomm3.CIPDriver, req_args):
    """
    Executes a function on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        req_args: The request to execute.  Typically a list like ['DLL Path','Function Name','Function Args'] 

    Returns:
        resp_code (int): The message response code.
        resp_data (str): The function response data.

    Raises:
        Exception: If the message response code is not zero, indicating an error.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->N-1  | Request Arguments                                   |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Response code (typically 0 = ran, other = error)    |
        | Bytes 4->N-1  | Function response                                   |
        | Bytes N       | Null footer                                         |

    Note:
        Only a certain subset of functions are whitelisted for access
        by this method.  They are documented in the HelperFunctions enum.
    """
    req_data = b''.join(arg.encode() + b'\x00' for arg in req_args)

    resp = messages.run_function(cip, req_data)
    if not resp: raise Exception(f'Failed to run function: {req_args}.')
    resp_code = int.from_bytes(resp.value[:4], byteorder='little', signed=False)
    resp_data = resp.value[4:].decode('utf-8').strip('\x00')
    return resp_code, resp_data

def create_directory(cip: pycomm3.CIPDriver, dir: str) -> bool:
    req_args = [helper_file_path, HelperFunctions.CREATE_DIRECTORY.value, dir]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != CREATE_DIR_SUCCESS): raise Exception('Failed to create directory on terminal.')    
    return True

def create_mer_list(cip: pycomm3.CIPDriver):
    req_args = [helper_file_path,HelperFunctions.CREATE_FILE_LIST.value,storage_path + '\\Rockwell Software\\RSViewME\\Runtime\\*.mer::' + upload_list_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return True

def create_runtime_directory(cip: pycomm3.CIPDriver, file: MEFile) -> bool:
    # Create paths
    if not(create_directory(cip, storage_path)): return False
    if not(create_directory(cip, storage_path + '\\Rockwell Software')): return False
    if not(create_directory(cip, storage_path + '\\Rockwell Software\\RSViewME')): return False
    if not(create_directory(cip, storage_path + '\\Rockwell Software\\RSViewME\\Runtime')): return False
    return True

def delete_file(cip: pycomm3.CIPDriver, file: str) -> bool:
    req_args = [helper_file_path,HelperFunctions.DELETE_FILE.value,file]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to delete file on remote terminal {file}')
    return True

def delete_file_mer_list(cip: pycomm3.CIPDriver) -> bool:
    return delete_file(cip, upload_list_path)

def get_file_exists(cip: pycomm3.CIPDriver, file: MEFile) -> bool:
    req_args = [helper_file_path, HelperFunctions.GET_FILE_EXISTS.value, storage_path + f'\\Rockwell Software\\RSViewME\\Runtime\\{file.name}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_file_size(cip: pycomm3.CIPDriver, file: MEFile) -> int:
    req_args = [helper_file_path, HelperFunctions.GET_FILE_SIZE.value, storage_path + f'\\Rockwell Software\\RSViewME\\Runtime\\{file.name}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return int(resp_data)

def get_folder_exists(cip: pycomm3.CIPDriver) -> bool:
    req_args = [helper_file_path, HelperFunctions.GET_FOLDER_EXISTS.value, storage_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0):
        warn(f'Response code was not zero.  Examine packets.')
        return False
    return bool(int(resp_data))

def get_free_space(cip: pycomm3.CIPDriver) -> int:
    req_args = [helper_file_path, HelperFunctions.GET_FREE_SPACE.value, storage_path + '\\Rockwell Software\\RSViewME\\Runtime\\']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return int(resp_data)

def get_helper_version(cip: pycomm3.CIPDriver) -> str:
    req_args = [helper_file_path, HelperFunctions.GET_VERSION.value, helper_file_path]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return str(resp_data)

def reboot(cip: pycomm3.CIPDriver):
    # For some reason this one has an extra trailing byte.
    # Not sure if it has some other purpose yet
    req_args = [helper_file_path, HelperFunctions.REBOOT.value,'']
    req_data = b''.join(arg.encode() + b'\x00' for arg in req_args)

    try:
        resp = messages.run_function(cip, req_data)
        
        #Should never get here
        raise Exception(resp)
    except pycomm3.exceptions.CommError as e:
        # Unlike most CIP messages, this one is always expected to
        # create an exception.  When it is received by the terminal,
        # the device reboots and breaks the socket.
        if (str(e) != 'failed to receive reply'): raise e

def set_startup_mer(cip: pycomm3.CIPDriver, file: MEFile, replace_comms: bool, delete_logs: bool) -> bool:
    req_args = [helper_file_path, HelperFunctions.CREATE_ME_SHORTCUT.value, storage_path + f':{file.name}: /r /delay']
    if replace_comms: req_args = [req_args[1], req_args[2], req_args[3] + ' /o']
    if delete_logs: req_args = [req_args[1], req_args[2], req_args[3] + ' /d']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return True


import os
import pycomm3

from . import terminal
from .types import *
from warnings import warn

class MEUtility(object):
    def __init__(self, comms_path: str, **kwargs):
        """
        Initializes an instance of the MEUtility class.

        Args:
            comms_path (str): The path to the communications resource (ex: 192.168.1.20).
            **kwargs: Additional keyword arguments. 

                - ignore_terminal_valid (bool): Optional; if set to True, 
                the instance will ignore terminal validation checks when
                performing uploads, downloads, etc.
                Defaults to False.
        """
        self.comms_path = comms_path
        self.ignore_terminal_valid = kwargs.get('ignore_terminal_valid', False)

    def download(self, file_path: str, **kwargs) -> MEResponse:
        """
        Downloads a *.MER file from the local device to the remote terminal.

        Args:
            file_path (str): The local path to the *.MER file (ex: C:\\YourFolder\\YourProgram.MER)
            **kwargs: Additional keyword arguments. 
                - delete_logs (bool): Optional; if set to True, will configure
                the terminal to delete logs at startup.  Defaults to false.
                - overwrite (bool): Optional; if set to True, will allow the
                download to overwrite an existing *.MER file on the terminal
                with the same file name.  Defaults to false.
                - replace_comms (bool): Optional; if set to True, will replace
                the terminal's communications setup with the one from the *.MER
                file being downloaded.  Defaults to false.
                - run_at_startup (bool): Optional; if set to True, will also set
                this *.MER file to be run at terminal startup and reboot the terminal
                now.  Defaults to True.  Note that this is a departure from the ME
                Transfer Utility, where it is not checked by default.
        """
        self.delete_logs = kwargs.get('delete_logs', False)
        self.overwrite = kwargs.get('overwrite', False)
        self.replace_comms = kwargs.get('replace_comms', False)
        self.run_at_startup = kwargs.get('run_at_startup', True)

        with pycomm3.CIPDriver(self.comms_path) as cip:
            file = MEFile(os.path.basename(file_path), self.overwrite, False, file_path)

            # Validate device at this communications path is a terminal of known version.
            self.device = terminal.validation.get_terminal_info(cip)
            if not(terminal.validation.is_terminal_valid(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')
                
            # Validate that all starting conditions for downnload to terminal are as expected
            if not(terminal.validation.is_download_valid(cip, self.device, file)): raise Exception('Download to terminal is invalid.')

            # Perform *.MER download to terminal
            if not(terminal.actions.download_mer_file(cip, self.device, file, self.run_at_startup, self.replace_comms, self.delete_logs)): raise Exception('Download to terminal failed.')

        return MEResponse(self.device, 'Success')

    def get_terminal_info(self) -> MEResponse:
        """
        If no upload or download are desired, where terminal info would typically be checked
        as a prerequisite, this function can be called to generate similar log entries to
        get information about the remote terminal.
        """
        with pycomm3.CIPDriver(self.comms_path) as cip:
            self.device = terminal.validation.get_terminal_info(cip)
            if not(terminal.validation.is_terminal_valid(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            terminal.actions.create_log(cip, self.device)

        return MEResponse(self.device, 'Success')

    def reboot(self) -> MEResponse:
        """
        Reboots the remote terminal now.
        """
        with pycomm3.CIPDriver(self.comms_path) as cip:
            self.device = terminal.validation.get_terminal_info(cip)

            if not(terminal.validation.is_terminal_valid(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            terminal.actions.reboot(cip)

        return MEResponse(self.device, 'Success')

    def upload(self, file_path: str, **kwargs) -> MEResponse:
        """
        Uploads a *.MER file from the remote terminal to the local device.

        Args:
            file_path (str): The local path to the *.MER file (ex: C:\\YourFolder\\YourProgram.MER)
            **kwargs: Additional keyword arguments. 
                - overwrite (bool): Optional; if set to True, will replace the
                file on the local device with the uploaded copy from the remote terminal.
                Defaults to False.
                - remote_file_name (str): Optional; if provided, is used to
                specify a different remote filename on the terminal than the local
                filename specified as part of file_path where it will end up.
        """
        file = MEFile(os.path.basename(file_path), False, False, file_path)
        self.remote_file_name = kwargs.get('remote_file_name', file.name)
        self.overwrite = kwargs.get('overwrite', False)
        rem_file = MEFile(self.remote_file_name,False,False,file_path)

        # Create upload folder if it doesn't exist yet
        if not(os.path.exists(os.path.dirname(file.path))): os.makedirs(os.path.dirname(file.path))

        # Check for existing *.MER
        if not(self.overwrite) and (os.path.exists(file.path)): raise Exception(f'File {file.name} already exists.  Use kwarg overwrite=True to overwrite existing local file from the remote terminal.')

        with pycomm3.CIPDriver(self.comms_path) as cip:
            # Validate device at this communications path is a terminal of known version.
            self.device = terminal.validation.get_terminal_info(cip)
            if not(terminal.validation.is_terminal_valid(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            # Perform *.MER upload from terminal
            if not(terminal.actions.upload_mer_file(cip, self.device, file, rem_file)): raise Exception('Upload from terminal failed.')
        
        return MEResponse(self.device, 'Success')

    def upload_all(self, file_path: str, **kwargs):
        """
        Uploads all *.MER files from the remote terminal to the local device.

        Args:
            file_path (str): The local path to the target directory (ex: C:\\YourFolder)
            **kwargs: Additional keyword arguments. 
                - overwrite (bool): Optional; if set to True, will replace the
                file on the local device with the uploaded copy from the remote terminal.
                Defaults to False.
        """
        self.overwrite = kwargs.get('overwrite', False)

        # Create upload folder if it doesn't exist yet
        if not(os.path.exists(file_path)): os.makedirs(os.path.dirname(file_path))

        with pycomm3.CIPDriver(self.comms_path) as cip:
            # Validate device at this communications path is a terminal of known version.
            self.device = terminal.validation.get_terminal_info(cip)
            if not(terminal.validation.is_terminal_valid(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            mer_list = terminal.actions.upload_mer_list(cip, self.device)
            
            for mer in mer_list:
                if len(mer) > 0:
                    mer_path = os.path.join(file_path, mer)
                    file = MEFile(os.path.basename(mer_path), self.overwrite, False, mer_path)

                    # Check for existing *.MER
                    if not(self.overwrite) and (os.path.exists(mer_path)): raise Exception(f'File {mer_path} already exists.  Use kwarg overwrite=True to overwrite existing local file from the remote terminal.')

                    # Perform *.MER upload from terminal
                    if not(terminal.actions.upload_mer_file(cip, self.device, file, file)): raise Exception('Upload from terminal failed.')

        return MEResponse(self.device, 'Success')


from enum import Enum
import pycomm3

# Known CIP class codes that aren't previously defined by pycomm3.
# Further investigation needed.
class CipClasses(Enum):
    FUNCTION = int.from_bytes(b'\x04\xFD', byteorder='big')
    REGISTRY = int.from_bytes(b'\x04\xFE', byteorder='big')
    FILE = int.from_bytes(b'\x04\xFF', byteorder='big')

# Known CIP service codes that aren't previously defined by pycomm3.
# Further investigation needed.
class CipServices(Enum):
    EXECUTE = int.from_bytes(b'\x50', byteorder='big')
    READ_REGISTRY = int.from_bytes(b'\x51', byteorder='big')
    WRITE_FILE = int.from_bytes(b'\x52', byteorder='big')
    READ_FILE = int.from_bytes(b'\x53', byteorder='big')

def create_file_exchange(cip: pycomm3.CIPDriver, data):
    return cip.generic_message(
        service=pycomm3.Services.create,
        class_code=CipClasses.FILE.value,
        instance=0x00,
        request_data=data,
        connected=False,
        route_path=None
    )

def delete_file_exchange(cip: pycomm3.CIPDriver, instance: int):
    return cip.generic_message(
        service=pycomm3.Services.delete,
        class_code=CipClasses.FILE.value,
        instance=instance,
        connected=False,
        route_path=None
    )

def read_file_chunk(cip: pycomm3.CIPDriver, instance: int, data):
    return cip.generic_message(
        service=CipServices.READ_FILE.value,
        class_code=CipClasses.FILE.value,
        instance=instance,
        request_data=data,
        connected=False,
        route_path=None
    )

def read_registry(cip: pycomm3.CIPDriver, data):
    return cip.generic_message(
        service=CipServices.READ_REGISTRY.value,
        class_code=CipClasses.REGISTRY.value,
        instance=0x00,
        request_data=data,
        connected=False,
        route_path=None
    )

def run_function(cip: pycomm3.CIPDriver, data):
    return cip.generic_message(
        service=CipServices.EXECUTE.value,
        class_code=CipClasses.FUNCTION.value,
        instance=0x00,
        request_data=data,
        connected=False,
        route_path=None
    )

def write_file_chunk(cip: pycomm3.CIPDriver, instance: int, data):
    return cip.generic_message(
        service=CipServices.WRITE_FILE.value,
        class_code=CipClasses.FILE.value,
        instance=instance,
        request_data=data,
        connected=False,
        route_path=None
    )

def get_attr_unk(cip: pycomm3.CIPDriver, data):
    return cip.generic_message(
        service=pycomm3.Services.get_attribute_single,
        class_code=CipClasses.FILE.value,
        instance=0x00,
        request_data=data,
        connected=False,
        route_path=None
    )

def set_attr_unk(cip: pycomm3.CIPDriver, data):
    return cip.generic_message(
        service=pycomm3.Services.set_attribute_single,
        class_code=CipClasses.FILE.value,
        instance=0x01,
        request_data=data,
        connected=False,
        route_path=None
    )